# Vision specific
from PIL import Image
from pycocotools.cocoeval import COCOeval

import numpy as np
import json
import pickle
import os
from tqdm import tqdm
from os.path import join

FPS = 30

def mkdir2(path):
    if not os.path.isdir(path):
        os.makedirs(path)
    return path

def imread(path, method='PIL'):
    return np.array(Image.open(path))

def gen_results(db, opts, filename):

    def ltrb2ltwh_(bboxes):
        if len(bboxes):
            if bboxes.ndim == 1:
                bboxes[2:] -= bboxes[:2]
            else:
                bboxes[:, 2:] -= bboxes[:, :2]
        return bboxes

    def ltrb2ltwh(bboxes):
        bboxes = bboxes.copy()
        return ltrb2ltwh_(bboxes)
        
    seqs = db.dataset['sequences']
    print('Merging and converting results')
    results_ccf = []
    miss = 0

    for sid, seq in enumerate(tqdm(seqs)):
        frame_list = [img for img in db.imgs.values() if img['sid'] == sid]
        results = pickle.load(open(join(opts.out_dir, seq + '.pkl'), 'rb'))
        results_parsed = results['results_parsed']
        timestamps = results['timestamps']

        tidx_p1 = 0
        for ii, img in enumerate(frame_list):
            # pred, gt association by time
            t = ii/FPS
            while tidx_p1 < len(timestamps) and timestamps[tidx_p1] <= t:
                tidx_p1 += 1
            if tidx_p1 == 0:
                # no output
                miss += 1
                bboxes, scores, labels  = [], [], []
                masks, tracks = None, None
            else:
                tidx = tidx_p1 - 1
                result = results_parsed[tidx]
                bboxes, scores, labels, masks = result[:4]
                if len(result) > 4:
                    tracks = result[4]
                else:
                    tracks = None
                
            # convert to coco fmt
            n = len(bboxes)
            if n:
                # important: must create a copy, it is used in subsequent frames
                bboxes_ltwh = ltrb2ltwh(bboxes)
            
            for i in range(n):
                result_dict = {
                    'image_id': int(img['id']),
                    'bbox': [float(a) for a in bboxes_ltwh[i]],
                    'score': float(scores[i]),
                    'category_id': int(labels[i]),
                }
                if masks is not None:
                    result_dict['segmentation'] = masks[i]
                results_ccf.append(result_dict)

    out_path = join(opts.out_dir, filename)
    if opts.overwrite or not isfile(out_path):
        json.dump(results_ccf, open(out_path, 'w'))

def evaluate(db, out_dir, filename, overwrite=False):
    results_ccf = join(out_dir, filename)
    eval_summary = eval_ccf(db, results_ccf)
    out_path = join(out_dir, 'eval_summary.pkl')
    if overwrite or not isfile(out_path):
        pickle.dump(eval_summary, open(out_path, 'wb'))

def eval_ccf(db, results, iou_type='bbox'):
    # ccf means CoCo Format
    if isinstance(results, str):
        if results.endswith('.pkl'):
            results = pickle.load(open(results, 'rb'))
        else:
            results = json.load(open(results, 'r'))
   
    results = db.loadRes(results)
    cocoEval = COCOeval(db, results, iou_type)
    cocoEval.evaluate()
    cocoEval.accumulate()
    cocoEval.summarize()

    return {
        'eval': cocoEval.eval,
        'stats': cocoEval.stats,
    }


import grpc

# classes generated by grpc
from .generated import eval_server_pb2
from .generated import eval_server_pb2_grpc

# other imports
import numpy as np 
import time
from time import perf_counter
from datetime import datetime

# multiprocessing 
import multiprocessing as mp
from multiprocessing import shared_memory
from threading import Thread


def get_grpc_result(bboxes, scores, labels):
    if len(bboxes) != len(scores) or len(bboxes) != len(labels):
        raise ValueError('number of bboxes, labels, bbox_scores must be same')
    res = eval_server_pb2.Result()
    for i in range(len(bboxes)):
        bbox = eval_server_pb2.Bbox()
        bbox.x1 = bboxes[i][0]
        bbox.y1 = bboxes[i][1]
        bbox.x2 = bboxes[i][2]
        bbox.y2 = bboxes[i][3]

        res.bboxes.append(bbox)
        res.bbox_scores.append(scores[i])
        res.labels.append(labels[i])
    return res

# receive input fidx streamed by server, store them in a list
def receive_stream(seq, latest_fidx, fid_ptr_dict, is_stream_ready, stream_start_time, config, verbose=False):
    if verbose:
        print("EvalClient (", datetime.now(), "): ", "Requesting stream for sequence ", seq)
    channel = grpc.insecure_channel(config['loopback_ip'] + ":" + str(config['image_service_port']))
    stub = eval_server_pb2_grpc.ImageServiceStub(channel)
    stream_request = eval_server_pb2.String(value=seq)
    send_times = []

    # receive input stream
    for i, response in enumerate(stub.GetImageStream(stream_request)):
        if i == 0:
            stream_start_time.value = perf_counter()
            if verbose:
                print("EvalClient (", datetime.now(), "): ", "Receiving stream for sequence ", seq) 
        if response.end_marker:
            latest_fidx.value = -1
            break
        is_stream_ready.clear()
        latest_fidx.value = response.fid
        fid_ptr_dict[response.fid] = (response.start_ptr, response.end_ptr)
        if response.fid >= 0:
            is_stream_ready.set()
        send_times.append(perf_counter() - response.timestamp)
    
    # print("EvalClient (", datetime.now(), "): ", "Mean sending time = ", np.mean(send_times), "s, stdev = ", np.std(send_times))
    # print("EvalClient (", datetime.now(), "): ", "Max/min sending time = ", np.max(send_times), np.min(send_times))
    # print("EvalClient (", datetime.now(), "): ", "Histogram = ", np.histogram(send_times))

    channel.close()

class EvalClient:

    def __init__(self, config, state=None, verbose=False):
        self.img_width, self.img_height = 1920, 1200

        if state is None:
            mp.set_start_method('spawn')
            self.latest_fidx = mp.Value('i', -1, lock=True)
            self.is_stream_ready = mp.Event()
            self.fid_ptr_dict = mp.Manager().dict()
            self.stream_start_time = mp.Value('d', 0.0, lock=True)
        else:
            self.latest_fidx = state[0]
            self.is_stream_ready = state[1]
            self.fid_ptr_dict = state[2]

        self.verbose = verbose
        # create image receiver stub
        self.channel = grpc.insecure_channel(config['loopback_ip'] + ":" + str(config['image_service_port']))
        self.config = config
        self.stub = eval_server_pb2_grpc.ImageServiceStub(self.channel)
        response = self.stub.GetShm(eval_server_pb2.Empty())
        self.existing_shm = shared_memory.SharedMemory(name=response.value)
        self.channel.close()

        # create result sender stub
        self.result_channel = grpc.insecure_channel(config['loopback_ip'] + ":" + str(config['result_service_port']))
        self.result_stub = eval_server_pb2_grpc.ResultServiceStub(self.result_channel)

        self.is_stream_ready.clear()
        self.stream_process = None

        self.latest_grpc_result = None
        self.grpc_result_ready = mp.Event()
        self.grpc_result_ready.clear()
        self.sequence_ended = False
        self.res_thread = None

    def get_state(self):
        return (self.latest_fidx, self.is_stream_ready, self.fid_ptr_dict)

    def close(self, results_file='results.json'):
        self.result_stub.GenResults(eval_server_pb2.String(value=results_file))
        self.result_channel.close()

    def result_stream_iterator(self):
        while 1:
            self.grpc_result_ready.wait()
            if self.sequence_ended:
                break
            yield self.latest_grpc_result
            self.grpc_result_ready.clear()

    def send_result_stream(self):
        result_iterator = self.result_stream_iterator()
        response = self.result_stub.PutResultStream(result_iterator)

    def stop_stream(self):
        self.stream_process.join()
        self.result_stub.FinishSequence(eval_server_pb2.Empty())
        self.is_stream_ready.clear()

        self.stream_process = None
        self.sequence_ended = True
        self.grpc_result_ready.set()
        self.res_thread.join()

    def request_stream(self, seq):
        self.sequence_ended = False
        self.grpc_result_ready.clear()
        self.res_thread = Thread(target=self.send_result_stream)
        self.res_thread.start()
        
        # fid_stream_receiver as processs
        self.stream_process = mp.Process(target=receive_stream, args=(seq, self.latest_fidx, self.fid_ptr_dict, self.is_stream_ready, self.stream_start_time, self.config, self.verbose))
        self.stream_process.start()

    def get_latest_fidx(self):
        self.is_stream_ready.wait()
        return self.latest_fidx.value

    def get_frame(self, fid=None, ptr=False):
        if fid is not None and fid < 0:
            raise TypeError(f"fid must be non-negative")
        if fid is None:
            fid = self.get_latest_fidx()
            if fid == -1:
                return None, None
        elif fid not in self.fid_ptr_dict:
            raise KeyError(f"frame not available yet")
        start_ptr, end_ptr = self.fid_ptr_dict[fid]
        if ptr:
            return fid, int(start_ptr/(self.img_height*self.img_width*3))
        return fid, np.ndarray((self.img_height, self.img_width, 3), dtype=np.uint8, buffer=self.existing_shm.buf[start_ptr:end_ptr])
    
    def send_result_async(self, bboxes, bbox_scores, labels):
        timestamp = perf_counter()
        grpc_result = get_grpc_result(bboxes, bbox_scores, labels)
        grpc_result.timestamp = timestamp
        self.latest_grpc_result = grpc_result
        self.grpc_result_ready.set()

    def send_result_to_server(self, bboxes, bbox_scores, labels):
        a = Thread(target=self.send_result_async, args=(bboxes, bbox_scores, labels))
        a.start()

    def get_frame_buf(self):
        return self.existing_shm

    def get_stream_start_time(self):
        self.is_stream_ready.wait()
        return self.stream_start_time.value


# gRPC imports
import grpc
from concurrent import futures

# classes generated by grpc
from .generated import eval_server_pb2
from .generated import eval_server_pb2_grpc

# Other imports
import numpy as np
import argparse
import pickle
from time import perf_counter
from os.path import join, isfile
from datetime import datetime
import os
import cmd
import time
import json

# multiprocessing
import multiprocessing as mp
from multiprocessing import shared_memory
from pycocotools.coco import COCO

# utils
from .utils import imread, gen_results, evaluate, mkdir2

FPS = 30

def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--data-root', type=str, required=True)
    parser.add_argument('--annot-path', type=str, required=True)
    parser.add_argument('--eval-config', type=str, required=False)
    parser.add_argument('--out-dir', type=str, required=True)
    parser.add_argument('--perf-factor', type=float, default=1)
    parser.add_argument('--overwrite', action='store_true', default=False)
    parser.add_argument('--log', type=str, default="server_log.txt")

    opts = parser.parse_args()
    return opts

class ImageServiceServicer(eval_server_pb2_grpc.ImageServiceServicer):

    def __init__(self, opts, db, current_sid, sequence_start_times_dict, log):
        self.opts = opts
        self.db = db
        self.seqs = self.db.dataset['sequences']
        self.seq_dirs = self.db.dataset['seq_dirs']
        self.fps = FPS

        img = self.db.imgs[0]
        self.img_width, self.img_height = img['width'], img['height']
        self.max_seq_len = 1000
        self.current_sid = current_sid
        self.sequence_start_times_dict = sequence_start_times_dict
        self.log = log

        # calculate the image size for shared memory allocation
        if 'max_seq_len' in self.db.dataset:
            self.max_seq_len = self.db.dataset['max_seq_len']
        else:
            seq_len = [len([img for img in self.db.imgs.values() if img['sid'] == sid])
                for sid in range(len(self.seqs))]
            self.max_seq_len = max(seq_len)

        self.shm = shared_memory.SharedMemory(create=True, size=(self.max_seq_len+2)*self.img_height*self.img_width*3)

    def close(self):
        self.shm.close()
        self.shm.unlink()

    def GetImageStream(self, request, context):
        
        n_total = 0
        for sid, seq in enumerate(self.seqs):
            if seq != request.value:
                continue
            print(datetime.now(), ": ", "Serving request - sid: " + str(sid) + " of " + str(len(self.seqs)-1), " name: ", request.value, file=self.log)
            frame_list = [img for img in self.db.imgs.values() if img['sid'] == sid]
    
            # load all frames in advance
            frames_local = np.ndarray((self.max_seq_len,) + (self.img_height, self.img_width, 3), dtype=np.uint8)
            frames = np.ndarray((self.max_seq_len,) + (self.img_height, self.img_width, 3), dtype=np.uint8, buffer=self.shm.buf)

            frames_permuted = np.random.permutation(len(frame_list))

            for i, img in enumerate(frame_list):
                if i % int(len(frame_list)/10) == 0:
                    print(datetime.now(), ": ", "Loaded frames (" + str(10*i/int(len(frame_list)/10)) + "%)", file=self.log)
                img_path = join(self.opts.data_root, self.seq_dirs[sid], img['name'])
                frames_local[frames_permuted[i]] = imread(img_path)
            frames[:] = frames_local[:]

            n_frame = len(frame_list)
            n_total += n_frame
            
            last_fidx = None
            
            t_total = (n_frame)/self.fps
            t_start = perf_counter()
            
            frames_processed = 0
            self.current_sid.value = sid
            self.sequence_start_times_dict[sid] = perf_counter()

            while 1:

                t1 = perf_counter()
                t_elapsed = t1 - t_start
                if t_elapsed >= t_total:
                    break

                # identify latest available frame
                fidx = int(np.floor(t_elapsed*self.fps))
                if fidx == last_fidx:
                    continue
                    
                frames_processed += 1
                last_fidx = fidx
                
                yield eval_server_pb2.Frame(fid = fidx, sid = sid, start_ptr=frames_permuted[fidx]*self.img_width*self.img_height*3, \
                                            end_ptr=(frames_permuted[fidx]+1)*self.img_width*self.img_height*3, timestamp=perf_counter())


            yield eval_server_pb2.Frame(end_marker=True, timestamp=perf_counter())
        print(datetime.now(), ": ","Finished streaming sequence ", request.value, file=self.log)

    def GetShm(self, request, context):
        return eval_server_pb2.String(value=self.shm.name)

class ResultServiceServicer(eval_server_pb2_grpc.ResultServiceServicer):

    def __init__(self, opts, db, current_sid, sequence_start_times_dict, log):
        self.opts = opts
        self.db = db
        self.seqs = self.db.dataset['sequences']
        self.sid_to_seq = {k:v for k,v in enumerate(self.seqs)}

        self.current_sid = current_sid
        self.sequence_start_times_dict = sequence_start_times_dict

        self.timestamps = []
        self.results_parsed = []
        self.input_fidx = []
        self.time_rcv = []
        self.log = log

        mkdir2(self.opts.out_dir)
    
    def reset_state(self):
        self.timestamps = []
        self.results_parsed = []
        self.input_fidx = []
        self.time_rcv = []

    def FinishSequence(self, result, context):
        seq_cur = self.sid_to_seq[self.current_sid.value]
        print(datetime.now(), ": ", "Writing result for sequence ", seq_cur, file=self.log)
        out_path = join(self.opts.out_dir, seq_cur + '.pkl')
        if self.opts.overwrite or not isfile(out_path):
            pickle.dump({
                'results_parsed': self.results_parsed,
                'timestamps': self.timestamps
            }, open(out_path, 'wb'))

        print(datetime.now(), ": ", "Mean time receive = ", np.mean(self.time_rcv), ", stdev = ", np.std(self.time_rcv), ", min = ", np.min(self.time_rcv), ", max = ", np.max(self.time_rcv), file=self.log)
        self.reset_state()
        return eval_server_pb2.Empty()

    def PutResultStream(self, result_iterator, context):
        for result in result_iterator:
            self.timestamps.append((perf_counter() - self.sequence_start_times_dict[self.current_sid.value])/self.opts.perf_factor)
            self.time_rcv.append(perf_counter() - result.timestamp)
            bboxes = np.ndarray((len(result.bboxes), 4), dtype=np.float32)
            bbox_scores = np.ndarray((len(result.bbox_scores),), dtype=np.float32)
            labels = np.ndarray((len(result.labels),), dtype=np.int64)
            for i in range(len(result.bboxes)):
                bboxes[i][0] = result.bboxes[i].x1
                bboxes[i][1] = result.bboxes[i].y1
                bboxes[i][2] = result.bboxes[i].x2
                bboxes[i][3] = result.bboxes[i].y2
                bbox_scores[i] = result.bbox_scores[i]
                labels[i] = result.labels[i]
            self.results_parsed.append((bboxes, bbox_scores, labels, None))
        return eval_server_pb2.Empty()

    def GenResults(self, request, context):
        gen_results(self.db, self.opts, request.value)
        return eval_server_pb2.Empty()

def result_server(opts, db, current_sid, sequence_start_times, res_rcv, log):
    print('Starting ResultService RPC Server. Listening on port ' + str(config['result_service_port']))

    result_server = grpc.server(futures.ThreadPoolExecutor(max_workers=12))
    result_servicer = ResultServiceServicer(opts, db, current_sid, sequence_start_times, log)
    eval_server_pb2_grpc.add_ResultServiceServicer_to_server(result_servicer, result_server)
    result_server.add_insecure_port('[::]:' + str(config['result_service_port']))
    result_server.start()
    msg = res_rcv.recv()
    if msg == 'terminate':
        result_server.stop(None)

class EvalShell(cmd.Cmd):
    intro = 'Welcome to the eval server.   Type help or ? to list commands.\n'
    prompt = '(eval_server) '

    def __init__(self, img_server, img_servicer, opts, db, res_send):
        super(EvalShell, self).__init__()
        self.db = db
        self.img_server = img_server
        self.img_servicer = img_servicer
        self.opts = opts
        self.res_send = res_send

    def do_close(self, arg):
        'Terminate evaluation server'
        self.img_servicer.close()
        self.img_server.stop(None)
        self.res_send.send('terminate')
        return True

    def do_gen_results(self, arg):
        'Generate and save results in coco-format to output directory. Usage: gen_results result_file.json'
        if not arg.endswith('.json') and not arg.endswith('.pkl'):
            print("Can only output to .json or .pkl file")
            return
        gen_results(self.db, self.opts, arg)
        print("Wrote results to: " + arg)

    def do_evaluate(self, arg):
        'Evaluate and display result metrics.'
        evaluate(self.db, self.opts.out_dir, arg, self.opts.overwrite)

    def do_log(self, arg):
        'View server log'
        os.system("cat " + self.opts.log)

if __name__ == '__main__':
    # create a gRPC server
    opts = parse_args()
    db = COCO(opts.annot_path)
    if opts.eval_config is not None:
        config = json.load(open(opts.eval_config, 'r'))
    else:
        config = {"image_service_port": 5007, "result_service_port": 5008, "loopback_ip": "127.0.0.1"}
        
    with open(opts.log, 'w+', 1) as log:
        # create dictionary to store current sequence and start times for each sequence stream
        current_sid = mp.Value('i', 0)
        sequence_start_times_dict = mp.Manager().dict()

        # start the image service RPC server
        server = grpc.server(futures.ThreadPoolExecutor(max_workers=12))
        servicer = ImageServiceServicer(opts, db, current_sid, sequence_start_times_dict, log)
        eval_server_pb2_grpc.add_ImageServiceServicer_to_server(servicer, server)
        server.add_insecure_port('[::]:' + str(config['image_service_port']))

        print('Starting ImageService RPC Server. Listening on port ' + str(config['image_service_port']))

        # start result service RPC server
        res_rcv, res_send = mp.Pipe(False)
        res_server = mp.Process(target=result_server, args=(opts, db, current_sid, sequence_start_times_dict, res_rcv, log))
        res_server.start()

        server.start()
        time.sleep(1)
        EvalShell(server, servicer, opts, db, res_send).cmdloop()
        server.wait_for_termination()







# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import eval_server_pb2 as eval__server__pb2


class ImageServiceStub(object):
    """service
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetImageStream = channel.unary_stream(
                '/ImageService/GetImageStream',
                request_serializer=eval__server__pb2.String.SerializeToString,
                response_deserializer=eval__server__pb2.Frame.FromString,
                )
        self.GetShm = channel.unary_unary(
                '/ImageService/GetShm',
                request_serializer=eval__server__pb2.Empty.SerializeToString,
                response_deserializer=eval__server__pb2.String.FromString,
                )


class ImageServiceServicer(object):
    """service
    """

    def GetImageStream(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetShm(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ImageServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetImageStream': grpc.unary_stream_rpc_method_handler(
                    servicer.GetImageStream,
                    request_deserializer=eval__server__pb2.String.FromString,
                    response_serializer=eval__server__pb2.Frame.SerializeToString,
            ),
            'GetShm': grpc.unary_unary_rpc_method_handler(
                    servicer.GetShm,
                    request_deserializer=eval__server__pb2.Empty.FromString,
                    response_serializer=eval__server__pb2.String.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'ImageService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ImageService(object):
    """service
    """

    @staticmethod
    def GetImageStream(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/ImageService/GetImageStream',
            eval__server__pb2.String.SerializeToString,
            eval__server__pb2.Frame.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetShm(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ImageService/GetShm',
            eval__server__pb2.Empty.SerializeToString,
            eval__server__pb2.String.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)


class ResultServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.PutResultStream = channel.stream_unary(
                '/ResultService/PutResultStream',
                request_serializer=eval__server__pb2.Result.SerializeToString,
                response_deserializer=eval__server__pb2.Empty.FromString,
                )
        self.FinishSequence = channel.unary_unary(
                '/ResultService/FinishSequence',
                request_serializer=eval__server__pb2.Empty.SerializeToString,
                response_deserializer=eval__server__pb2.Empty.FromString,
                )
        self.GenResults = channel.unary_unary(
                '/ResultService/GenResults',
                request_serializer=eval__server__pb2.String.SerializeToString,
                response_deserializer=eval__server__pb2.Empty.FromString,
                )


class ResultServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def PutResultStream(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FinishSequence(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GenResults(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ResultServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'PutResultStream': grpc.stream_unary_rpc_method_handler(
                    servicer.PutResultStream,
                    request_deserializer=eval__server__pb2.Result.FromString,
                    response_serializer=eval__server__pb2.Empty.SerializeToString,
            ),
            'FinishSequence': grpc.unary_unary_rpc_method_handler(
                    servicer.FinishSequence,
                    request_deserializer=eval__server__pb2.Empty.FromString,
                    response_serializer=eval__server__pb2.Empty.SerializeToString,
            ),
            'GenResults': grpc.unary_unary_rpc_method_handler(
                    servicer.GenResults,
                    request_deserializer=eval__server__pb2.String.FromString,
                    response_serializer=eval__server__pb2.Empty.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'ResultService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ResultService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def PutResultStream(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/ResultService/PutResultStream',
            eval__server__pb2.Result.SerializeToString,
            eval__server__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def FinishSequence(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ResultService/FinishSequence',
            eval__server__pb2.Empty.SerializeToString,
            eval__server__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GenResults(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ResultService/GenResults',
            eval__server__pb2.String.SerializeToString,
            eval__server__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name = "sap-toolkit",
    version = "0.0.4",
    url = "https://github.com/karthiksharma98/sap-starterkit/tree/master/sap-toolkit",
    author = "Kartikeya Sharma",
    author_email = "ksharma@illinois.edu",
    classifiers =
        ["Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"],
    description = "Toolkit to benchmark and evaluate streaming perception algorithms",
    packages = setuptools.find_packages(),
    long_description = long_description,
    long_description_content_type = "text/markdown",
    python_requires='>=3.8',
)


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: eval_server.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='eval_server.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x11\x65val_server.proto\"l\n\x05\x46rame\x12\x0b\n\x03\x66id\x18\x01 \x01(\x05\x12\x11\n\tstart_ptr\x18\x02 \x01(\x03\x12\x0f\n\x07\x65nd_ptr\x18\x03 \x01(\x03\x12\x0b\n\x03sid\x18\x04 \x01(\x05\x12\x12\n\nend_marker\x18\x06 \x01(\x08\x12\x11\n\ttimestamp\x18\x07 \x01(\x01\"\x17\n\x06String\x12\r\n\x05value\x18\x01 \x01(\t\"6\n\x04\x42\x62ox\x12\n\n\x02x1\x18\x01 \x01(\x01\x12\n\n\x02y1\x18\x02 \x01(\x01\x12\n\n\x02x2\x18\x03 \x01(\x01\x12\n\n\x02y2\x18\x04 \x01(\x01\"d\n\x06Result\x12\x15\n\x06\x62\x62oxes\x18\x01 \x03(\x0b\x32\x05.Bbox\x12\x13\n\x0b\x62\x62ox_scores\x18\x02 \x03(\x01\x12\x0e\n\x06labels\x18\x03 \x03(\x05\x12\x0b\n\x03\x66id\x18\x04 \x01(\x03\x12\x11\n\ttimestamp\x18\x05 \x01(\x01\"\x07\n\x05\x45mpty2R\n\x0cImageService\x12%\n\x0eGetImageStream\x12\x07.String\x1a\x06.Frame\"\x00\x30\x01\x12\x1b\n\x06GetShm\x12\x06.Empty\x1a\x07.String\"\x00\x32|\n\rResultService\x12&\n\x0fPutResultStream\x12\x07.Result\x1a\x06.Empty\"\x00(\x01\x12\"\n\x0e\x46inishSequence\x12\x06.Empty\x1a\x06.Empty\"\x00\x12\x1f\n\nGenResults\x12\x07.String\x1a\x06.Empty\"\x00\x62\x06proto3'
)




_FRAME = _descriptor.Descriptor(
  name='Frame',
  full_name='Frame',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='fid', full_name='Frame.fid', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_ptr', full_name='Frame.start_ptr', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_ptr', full_name='Frame.end_ptr', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sid', full_name='Frame.sid', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_marker', full_name='Frame.end_marker', index=4,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='Frame.timestamp', index=5,
      number=7, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21,
  serialized_end=129,
)


_STRING = _descriptor.Descriptor(
  name='String',
  full_name='String',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='String.value', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=131,
  serialized_end=154,
)


_BBOX = _descriptor.Descriptor(
  name='Bbox',
  full_name='Bbox',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='x1', full_name='Bbox.x1', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='y1', full_name='Bbox.y1', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='x2', full_name='Bbox.x2', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='y2', full_name='Bbox.y2', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=156,
  serialized_end=210,
)


_RESULT = _descriptor.Descriptor(
  name='Result',
  full_name='Result',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='bboxes', full_name='Result.bboxes', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bbox_scores', full_name='Result.bbox_scores', index=1,
      number=2, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='labels', full_name='Result.labels', index=2,
      number=3, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fid', full_name='Result.fid', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='Result.timestamp', index=4,
      number=5, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=212,
  serialized_end=312,
)


_EMPTY = _descriptor.Descriptor(
  name='Empty',
  full_name='Empty',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=314,
  serialized_end=321,
)

_RESULT.fields_by_name['bboxes'].message_type = _BBOX
DESCRIPTOR.message_types_by_name['Frame'] = _FRAME
DESCRIPTOR.message_types_by_name['String'] = _STRING
DESCRIPTOR.message_types_by_name['Bbox'] = _BBOX
DESCRIPTOR.message_types_by_name['Result'] = _RESULT
DESCRIPTOR.message_types_by_name['Empty'] = _EMPTY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Frame = _reflection.GeneratedProtocolMessageType('Frame', (_message.Message,), {
  'DESCRIPTOR' : _FRAME,
  '__module__' : 'eval_server_pb2'
  # @@protoc_insertion_point(class_scope:Frame)
  })
_sym_db.RegisterMessage(Frame)

String = _reflection.GeneratedProtocolMessageType('String', (_message.Message,), {
  'DESCRIPTOR' : _STRING,
  '__module__' : 'eval_server_pb2'
  # @@protoc_insertion_point(class_scope:String)
  })
_sym_db.RegisterMessage(String)

Bbox = _reflection.GeneratedProtocolMessageType('Bbox', (_message.Message,), {
  'DESCRIPTOR' : _BBOX,
  '__module__' : 'eval_server_pb2'
  # @@protoc_insertion_point(class_scope:Bbox)
  })
_sym_db.RegisterMessage(Bbox)

Result = _reflection.GeneratedProtocolMessageType('Result', (_message.Message,), {
  'DESCRIPTOR' : _RESULT,
  '__module__' : 'eval_server_pb2'
  # @@protoc_insertion_point(class_scope:Result)
  })
_sym_db.RegisterMessage(Result)

Empty = _reflection.GeneratedProtocolMessageType('Empty', (_message.Message,), {
  'DESCRIPTOR' : _EMPTY,
  '__module__' : 'eval_server_pb2'
  # @@protoc_insertion_point(class_scope:Empty)
  })
_sym_db.RegisterMessage(Empty)



_IMAGESERVICE = _descriptor.ServiceDescriptor(
  name='ImageService',
  full_name='ImageService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=323,
  serialized_end=405,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetImageStream',
    full_name='ImageService.GetImageStream',
    index=0,
    containing_service=None,
    input_type=_STRING,
    output_type=_FRAME,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetShm',
    full_name='ImageService.GetShm',
    index=1,
    containing_service=None,
    input_type=_EMPTY,
    output_type=_STRING,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_IMAGESERVICE)

DESCRIPTOR.services_by_name['ImageService'] = _IMAGESERVICE


_RESULTSERVICE = _descriptor.ServiceDescriptor(
  name='ResultService',
  full_name='ResultService',
  file=DESCRIPTOR,
  index=1,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=407,
  serialized_end=531,
  methods=[
  _descriptor.MethodDescriptor(
    name='PutResultStream',
    full_name='ResultService.PutResultStream',
    index=0,
    containing_service=None,
    input_type=_RESULT,
    output_type=_EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='FinishSequence',
    full_name='ResultService.FinishSequence',
    index=1,
    containing_service=None,
    input_type=_EMPTY,
    output_type=_EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GenResults',
    full_name='ResultService.GenResults',
    index=2,
    containing_service=None,
    input_type=_STRING,
    output_type=_EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_RESULTSERVICE)

DESCRIPTOR.services_by_name['ResultService'] = _RESULTSERVICE

# @@protoc_insertion_point(module_scope)
